### PersefoneDesktop - a GUI to the Persefone model of agriculture and ecosystems
###
### This file manages the model initialisation and run functions.
###

@resumable function createlaunchfunction()
    global model, landcovermap
    running[] = false
    progress[] = 0.0
    @yield 1
    model = initialise(configfile[])
    @yield 1
    model.date -= Day(1)
    Persefone.outputdata(model, true) # record data before run starts
    model.date += Day(1)
    landcovermap = rotr90(load(joinpath(@param(world.mapdirectory), @param(world.landcovermap))))
    date[] = model.date
    launching[] = false
end

function nextstep()
    global model
    if date[] >= @param(core.enddate)
        running[] = false
        runbuttontext[] = ">>"
        runbuttontip[] = "Run"
        return
    elseif date[] < model.date
        # If we've "scrolled back" in time, step forward again without
        # having to resimulate (see `previousstep()`)
        date[] += Day(1)
    else
        # If we're already displaying the newest model update, we need to
        # keep simulating
        stepsimulation!(model)
        date[] = model.date
    end
    progress[] = (date[]-@param(core.startdate)) /
        (@param(core.enddate)-@param(core.startdate))
    @emit updateMakie()
    println("Updated model: $(date[]-Day(1))")
end

#XXX This could also be done with Channels rather than ResumableFunctions,
# which would save us one dependency (Channels are inbuilt). However, the
# ResumableFunctions version seemed to work more smoothly.
@resumable function createrunfunction()
    i = round(delay[]*20)
    while running[]
        i -= 1
        if i <= 0
            nextstep()
            i = round(delay[]*20)
        end
        # to reduce interface lag, always only sleep for 50ms
        sleep(0.05)
        @yield 1
    end
end

function previousstep()
    # Since we store all model output in a dataframe, we can simply
    # "scroll back" in time and visualise the state of the model at
    # a previous date
    global model
    date[] <= @param(core.startdate) && return
    date[] -= Day(1)
    progress[] = (date[]-@param(core.startdate)) /
        (@param(core.enddate)-@param(core.startdate))
    @emit updateMakie()
    println("Backtracked model: $(date[]-Day(1))")
end

on(launching; priority=10) do l
    if l
        QML.start(timer)
        global newsimulation = createlaunchfunction()
        println("Setting up model.") 
    else
        QML.stop(timer)
        println("Model setup complete.")
    end
end

on(running; priority=10) do r
    if r
        QML.start(timer)
        global runsimulation = createrunfunction()
        println("Simulation started.")
    else
        !launching[] && QML.stop(timer)
        println("Simulation stopped.")
    end
end

on(ticks) do t
    launching[] && newsimulation()
    running[] && runsimulation()
    t >= typemax(Int32) && (ticks[] = 0) #prevent overflow
end
